// ============================================================================
// 
// Copyright (c) 1998 The CGAL Consortium
//
// This software and related documentation is part of an INTERNAL release
// of the Computational Geometry Algorithms Library (CGAL). It is not
// intended for general use.
//
// ----------------------------------------------------------------------------
//
// release       : $CGAL_Revision: CGAL-1.0$
// release_date  : $CGAL_Date: 04/05/98 $
//
// file          : include/CGAL/Pm_for_bops_overlay.h
// source        : ppau/project/CGAL/progs/ovl/ovl-for-bops1.2/Pm_for_bops_overlay.h
// revision      : $Revision:  $
// revision_date : $ $
// author(s)     : Petru Pau <Petru.Pau@risc.uni-linz.ac.at>
//
// coordinator   : RISC Linz (Sabine Stifter <stifter@cyber.at>)
//
// 
// ============================================================================

#ifndef CGAL_PM_OVERLAY_H
#define CGAL_PM_OVERLAY_H

#ifndef CGAL_PREDICATES_ON_POINTS_2
#include <CGAL/predicates_on_points_2.h>
#endif

#ifndef SEGMENT_2_SEGMENT_2_INTERSECTION_H
#include <CGAL/Segment_2_Segment_2_intersection.h>
#endif

#ifndef CGAL_PM_DEFAULT_DCEL_H
#include <CGAL/Pm_default_dcel.h>
#endif
 
#ifndef CGAL_PLANAR_MAP_2_H
#include <CGAL/Planar_map_2.h>
#endif

#include <list.h>
#include <set.h>

//#define DEBUG_COLLECT_PORTIONS
//#define DEBUG_INSERT_EDGES
#define DEBUG_COLORIZE

enum Color{
    NO_COLOR 	= 0,
    RED		= 1,
    BLACK 	= 2,
    RED_AND_BLACK = 3,
    UNCOLORED   = -1
};

/**************************************************************
************* B O P S _ T R A I T S   C L A S S ***************
**************************************************************/
template <class Traits>
class Bops_traits : public Traits
{
public:
  struct Info_edge
    {
      Color map_color;
      Color edge_color;
      Color left_face_color;
  };
  
  typedef Color Info_face;

  typedef Color Info_vertex;
  
};

/**************************************************************
********* P O I N T   F O R   B O P S   C L A S S *************
*********                                         *************
******** used to store all the intersection points ************
******** of an edge with edges in the other planar ************
******** map                                       ************
**************************************************************/
template<class T>
class Point_for_pm_bops : public T
{
  // If $p$ is a NORMAL point, no color is assigned;
  // If $p$ is a BEGIN_COMMON point, its color is the color of
       // the left face of the edge from the other planar map;
  // If $p$ is a END_COMMON point, its color is the color of
       // the right face of the edge from the other planar map.
public:
  enum Point_type 
  {
    NORMAL_POINT = 0, 
    BEGIN_COMMON = 1, 
    END_COMMON   = 2
  };
  Point_for_pm_bops(T t) : T(t) {
      _color = NO_COLOR;
      _type  = NORMAL_POINT;
    }
  Point_for_pm_bops(T p, Point_type t, Color c) : T(p) {
      _color = c;
      _type  = t;
    }
  Point_type type()  const { return _type; }
  Color      color() const { return _color; }
  void 	     print() const 
    {
#ifdef CGAL_CARTESIAN_H
      cout << hx() << " " << hy() << " " << hw() ;
#else // CGAL_HOMOGENOUS_H
      cout << x()  << " " << y();
#endif
      switch(_type){
      case NORMAL_POINT: cout << ", normal intersection, ";
	break;
      case BEGIN_COMMON: cout << ", begin common segment, ";
	break;
      case END_COMMON: cout << ", end common segment, ";
      }      
      switch(_color){
      case NO_COLOR: cout << "NO_COLOR\n"; 		break;
      case RED: cout << "RED\n"; 	   		break;
      case BLACK:cout << "BLACK\n";	   		break;
      case RED_AND_BLACK: cout << "RED AND BLACK\n"; 	break;
      case UNCOLORED: cout << "UNCOLORED! Imagine!\n";
      }
    }
      
private:
  Color _color;
  Point_type  _type;
};

template <class I>
struct CGAL_less_or_under: binary_function<I, I, bool>{
  bool operator () (const I& p1, const I& p2) const
  {
    if (CGAL_SMALLER == CGAL_compare_x(p1,p2)) return true;
    if (CGAL_LARGER  == CGAL_compare_x(p1,p2)) return false;
    if (CGAL_SMALLER == CGAL_compare_y(p1,p2)) return true;
    if (CGAL_LARGER  == CGAL_compare_y(p1,p2)) return false;
    if (I::NORMAL_POINT == p1.type() &&
	I::BEGIN_COMMON == p2.type()) return true;
    if (I::NORMAL_POINT == p2.type() &&
	I::BEGIN_COMMON == p1.type()) return false;
    if (I::NORMAL_POINT == p2.type() &&
	I::END_COMMON   == p1.type()) return true;
    if (I::NORMAL_POINT == p1.type() &&
	I::END_COMMON   == p2.type()) return false;
    if (I::END_COMMON   == p1.type() &&
	I::BEGIN_COMMON == p2.type()) return true;
    if (I::BEGIN_COMMON == p1.type() &&
	I::END_COMMON   == p2.type()) return false;
    return false;
  }
};

/**************************************************************
******** S E G M E N T   F O R   B O P S   C L A S S **********
*********                                          ************
******** used to store all the segments            ************ 
******** that will be inserted in the overlay      ************
**************************************************************/
template<class T>
class Segment_for_pm_bops : public T
{
  Color _color;
  Color _lf_color;
  Color _rf_color;
public:
  Color get_color() 	const { return _color; }
  Color get_lf_color() 	const { return _lf_color; }
  Color get_rf_color() 	const { return _rf_color; }
  Segment_for_pm_bops(T t, 
		      Color c, 
		      Color clf, 
		      Color crf) : T(t), 
                                   _color(c), 
                                   _lf_color(clf), 
                                   _rf_color(crf) { }
  
  ~Segment_for_pm_bops() { }
};

template <class I>
struct CGAL_before_or_under: binary_function<Segment_for_pm_bops<I>, 
			                     Segment_for_pm_bops<I>, 
			                     bool>{
  bool operator () (const Segment_for_pm_bops<I>& s1, 
		    const Segment_for_pm_bops<I>& s2) const
  {
    if (CGAL_SMALLER == CGAL_compare_x(s1.source(),s2.source())) return true;
    if (CGAL_LARGER  == CGAL_compare_x(s1.source(),s2.source())) return false;
    if (CGAL_SMALLER == CGAL_compare_y(s1.source(),s2.source())) return true;
    if (CGAL_LARGER  == CGAL_compare_y(s1.source(),s2.source())) return false;
    if (CGAL_SMALLER == CGAL_compare_x(s1.target(),s2.target())) return true;
    if (CGAL_LARGER  == CGAL_compare_x(s1.target(),s2.target())) return false;
    if (CGAL_SMALLER == CGAL_compare_y(s1.target(),s2.target())) return true;
    return false;
  }
};

/********************************************************************
*************** P M _ O V E R L A Y _ F O R _ B O P S ***************
***************                                       ***************
************** Computes the overlay and associates     **************
************** colors to its components                **************
********************************************************************/
template <class Traits>
class CGAL_Pm_overlay_for_bops: public Traits
{
public:
  typedef CGAL_Pm_default_dcel<Traits>    Dcel;
  typedef CGAL_Planar_map_2<Dcel, Traits> Planar_map;
  typedef CGAL_Planar_map_2<CGAL_Pm_default_dcel<Bops_traits<Traits> >, 
                                                 Bops_traits<Traits> >
                                          Planar_map_for_bops;

  typedef typename Planar_map::Halfedge_iterator 
    Edge_iterator;
  typedef typename Planar_map_for_bops::Vertex_iterator 
    Bops_vertex_iterator;
  typedef typename Planar_map_for_bops::Halfedge_iterator 
    Bops_edge_iterator;
  typedef typename Planar_map_for_bops::Face_iterator 
    Bops_face_iterator;

  typedef typename Traits::X_curve        Segment;
  typedef typename Traits::Point	  Point;
    
  typedef set<Segment_for_pm_bops<Segment>, 
              CGAL_before_or_under<Segment> > 
    Segment_container;
  typedef set<Point_for_pm_bops<Point>, 
              CGAL_less_or_under<Point_for_pm_bops<Point> > > 
    Point_container;

  CGAL_Pm_overlay_for_bops(const Planar_map& pm1, 
			   const Planar_map& pm2,
			         Planar_map_for_bops& pm_out)
    {
      Segment_container l_out;

      Collect_edges(pm1, pm2, l_out);
      Insert_edges(l_out, pm_out);
      Colorize(pm_out);
    }

  void Insert_edges(const Segment_container& all_segs, 
		    Planar_map_for_bops& pm)
  {
      typename Segment_container::const_iterator s_i;
      typename Planar_map_for_bops::Halfedge 	 new_edge;
      
      for (s_i=all_segs.begin(); s_i!=all_segs.end(); s_i++){

	new_edge = pm.insert(Segment((*s_i).source(), (*s_i).target()));
	if (new_edge.source().point() == (*s_i).source()){
	     typename Bops_traits<Traits>::Info_edge ie;
	     ie.map_color  = (*s_i).get_color() ;
	     ie.edge_color = (*s_i).get_color();
	     ie.left_face_color =  (*s_i).get_lf_color();
	  new_edge.set_info(ie);	
	     ie.left_face_color =  (*s_i).get_rf_color();
	  new_edge.twin().set_info(ie);
	}
	else
	  if (new_edge.target().point() == (*s_i).source()){
	     typename Bops_traits<Traits>::Info_edge ie;
	     ie.map_color  = (*s_i).get_color() ;
	     ie.edge_color = (*s_i).get_color();
	     ie.left_face_color =  (*s_i).get_rf_color();
	  new_edge.set_info( ie);	
	     ie.left_face_color =  (*s_i).get_lf_color();
	  new_edge.twin().set_info(ie);
         }
      }
      return;
    }

  void Colorize(Planar_map_for_bops& pm)
    {
      list<typename Planar_map_for_bops::Face> face_list;

      Bops_face_iterator 	fi;
      Bops_edge_iterator 	ei;
      Bops_vertex_iterator 	vi;
      
      typename Bops_traits<Traits>::Info_face this_face_color = UNCOLORED;
      typename Bops_traits<Traits>::Info_edge ie;
      typename Bops_traits<Traits>::Info_vertex iv;
      
      for(fi=pm.faces_begin(); fi!=pm.faces_end(); ++fi)
	(*fi).set_info(this_face_color);

      this_face_color = NO_COLOR;

      for(vi=pm.vertices_begin(); vi!=pm.vertices_end(); ++vi)
	(*vi).set_info(this_face_color);
      
      pm.unbounded_face().set_info(this_face_color);

      typename Planar_map_for_bops::Holes_iterator 
	hi=pm.unbounded_face().holes_begin();
      if (hi==pm.unbounded_face().holes_end()) return;

      typename Planar_map_for_bops::Ccb_halfedge_circulator ccb_i = *hi;
      typename Planar_map_for_bops::Halfedge 	 he = (*ccb_i).twin();
      typename Planar_map_for_bops::Face 	 f  = he.face();
      typename Planar_map_for_bops::Halfedge	 he_next = he.next_halfedge();
      typename Planar_map_for_bops::Vertex	 v;
      
      f.set_info(he.info().left_face_color);

      do{
	ie.edge_color     = (Color)((int)he_next.info().edge_color |
				    (int)he.info().left_face_color);
        ie.map_color      = he_next.info().map_color;
        ie.left_face_color= he_next.info().left_face_color;  

	he_next.set_info(ie);
        ie.left_face_color= he_next.twin().info().left_face_color;
	he_next.twin().set_info(ie);

	v = he_next.source();
	iv = (Color)((int)ie.edge_color | (int) v.info());
	v.set_info(iv);
	
	if (!he_next.twin().face().is_unbounded()){
	  Color tmp_color = NO_COLOR;
	  he_next.twin().face().set_info(tmp_color);
          face_list.push_back(he_next.twin().face());
	}
	
	he_next = he_next.next_halfedge();
      } while (he_next!=he);
  
      while (face_list.size()>0){
	f = face_list.front();
	face_list.pop_front();

	this_face_color = NO_COLOR;

	he = f.halfedge_on_outer_ccb();
		
	typename Planar_map_for_bops::Halfedge he_next = he;

	Color this_face_color = NO_COLOR;
	Color max_face_color  = NO_COLOR;

	bool has_both_colors = false;
	Color neighbor_to_unbounded = NO_COLOR;
	bool twins_face_changed = false;
	bool edge_adjacent_to_unbounded_face 	= false;
	
	int nr_colors = 0;
	
	typename Planar_map_for_bops::Halfedge he_neigh;

	do{
	  if (he_next.info().left_face_color == RED_AND_BLACK){
	    this_face_color = RED_AND_BLACK;
	    has_both_colors = true;
	  }
	  else{
	    if (he_next.twin().face().info() == NO_COLOR){
	      this_face_color = he_next.info().left_face_color;
	    }
	    else {
	      if (he_next.info().left_face_color == NO_COLOR){
		he_neigh = he_next;
		edge_adjacent_to_unbounded_face = true;
	      }
	      else {
		if (he_next.info().edge_color != he_next.info().map_color){
		  if (he_next.info().edge_color != (Color)3)
		    CGAL_kernel_assertion_msg(false, "Modified edge color != 3");
		  this_face_color = (Color)3;
		  twins_face_changed = true;
		}
	      }
	    }
	  }
	
	  max_face_color = (Color)((int)he_next.info().left_face_color | 
				   (int)max_face_color);
	  he_next = he_next.next_halfedge();
	} while (he_next!=he);
	
	
	if (has_both_colors || (max_face_color == RED_AND_BLACK)){
	  this_face_color = RED_AND_BLACK;
	  f.set_info(this_face_color);  
	}
	else
	  if (neighbor_to_unbounded != NO_COLOR){
	    f.set_info(neighbor_to_unbounded);  
	  }
	  else
	    if (edge_adjacent_to_unbounded_face){
	      if (max_face_color == 3)
		CGAL_kernel_assertion_msg(false, 
		     "Red_and_black face external to a map!");
	      f.set_info(max_face_color);
	    }
	  else
	    if (twins_face_changed) {
	      this_face_color = RED_AND_BLACK;
	      f.set_info(this_face_color);
	    }
	    else
	      f.set_info(max_face_color);
	
	do{
	  ie.edge_color     = (Color)((int)he_next.info().edge_color | 
				      (int)f.info());
	  ie.map_color      = he_next.info().map_color;
	  ie.left_face_color= he_next.info().left_face_color; 
	  
	  he_next.set_info(ie);

	  ie.left_face_color= he_next.twin().info().left_face_color;
	  he_next.twin().set_info(ie);

	  v = he_next.source();
	  iv = (Color)((int)ie.edge_color | (int) v.info());
	  v.set_info(iv);
	  
	  if (he_next.twin().face().info() == UNCOLORED){
	    Color tmp_color = NO_COLOR;
	    he_next.twin().face().set_info(tmp_color);
	    face_list.push_back(he_next.twin().face());
	  }
	  
	  he_next = he_next.next_halfedge();
          
	} while (he_next!=he);
      }

      return;
    }
  
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
//            S T A R T I N G    F R O M    H E R E                 //
//    these functions can be replaced with a faster version         //
//             eventually using the fast intersection               //
//                  of a set of segments                            //
//------------------------------------------------------------------//
  void Collect_edges(const Planar_map& pm1,
		     const Planar_map& pm2,
		           Segment_container& all_segs)
    {

      Edge_iterator e_i;
      Point p1, p2;
      Color e_color;
      Color lf_color;
      Color rf_color;
      
      for(e_i=pm1.halfedges_begin(); e_i!=pm1.halfedges_end(); ++e_i){
        Point_container o;
	p1 = (*e_i).source().point();
	p2 = (*e_i).target().point();

	if (CGAL_SMALLER == CGAL_compare_x(p1,p2) ||
            CGAL_x_equal(p1,p2) &&
            CGAL_SMALLER == CGAL_compare_y(p1, p2)) {
	  if ((*e_i).face().is_unbounded()) 
	    lf_color=NO_COLOR;
	  else
	    lf_color=RED;
	  if ((*e_i).twin().face().is_unbounded()) 
	    rf_color=NO_COLOR;
	  else
	    rf_color=RED;
	  e_color = RED;	  

	  Intersect_segment_planar_map(Segment(p1, p2), RED, pm2, o);
	  Collect_portions(p1, p2, e_color, lf_color, rf_color, o, all_segs);
	}
      }
      
     for(e_i=pm2.halfedges_begin(); e_i!=pm2.halfedges_end(); ++e_i){
        Point_container o;
        p1 = (*e_i).source().point();
	p2 = (*e_i).target().point();

	if (CGAL_SMALLER == CGAL_compare_x(p1,p2) ||
            CGAL_x_equal(p1,p2) &&
            CGAL_SMALLER == CGAL_compare_y(p1, p2)) {
	  if ((*e_i).face().is_unbounded()) 
	    lf_color=NO_COLOR;
	  else
	    lf_color=BLACK;
	  if ((*e_i).twin().face().is_unbounded()) 
	    rf_color=NO_COLOR;
	  else
	    rf_color=BLACK;
	  e_color = BLACK;
	  
	  Intersect_segment_planar_map(Segment(p1, p2), BLACK, pm1, o);
	  Collect_portions(p1, p2, e_color, lf_color, rf_color, o, all_segs);
	}
     }

     return;
    }
       
  void Intersect_segment_planar_map(const Segment& s, 
				    Color c,
				    const Planar_map& pm, 
				          Point_container& p_c)
    {
      Edge_iterator e_i;
      CGAL_Object o;
      Point p;
      Segment seg;

      for(e_i=pm.halfedges_begin(); e_i!=pm.halfedges_end(); e_i++){
	o=CGAL_intersection(s, (*e_i).curve());
	if (CGAL_assign(p, o))
	  p_c.insert(Point_for_pm_bops<Point>(p));
	if (CGAL_assign(seg,o)){
	  Point p1;
	  Point p2;
	  p1 = seg.source();
	  p2 = seg.target();
	  if (CGAL_SMALLER == CGAL_compare_x(p2,p1) ||
	      CGAL_x_equal(p1,p2) &&
	      CGAL_SMALLER == CGAL_compare_y(p2, p1)){
	    Point aux = p1;
	    p1 = p2;
	    p2 = aux;
	  }
	  Color s_col;
	  if ((*e_i).face().is_unbounded()) s_col = NO_COLOR;
	  else s_col = (Color)(3-(int)c);
  	  p_c.insert(Point_for_pm_bops<Point>(p1, 
		      Point_for_pm_bops<Point>::BEGIN_COMMON,
		      s_col));

	  if ((*e_i).twin().face().is_unbounded()) s_col = NO_COLOR;
	  else s_col = (Color)(3-(int)c);
	  p_c.insert(Point_for_pm_bops<Point>(p2, 
		     Point_for_pm_bops<Point>::END_COMMON,
		     s_col));
	}
      }

      return;
    }
  
  void Collect_portions(const Point& A,
			const Point& B,
			const Color& c,
			const Color& lf,
			const Color& rf,
			      Point_container& o,
			      Segment_container& all_segs)
    {
      typename Point_container::iterator pc_i;
      Point p1, p2;
      Color lf_color=lf, 
	    rf_color=rf, 
	    e_color =c;
      
      p1 = A;

      for(pc_i=o.begin(); pc_i!=o.end(); pc_i++)
      {
#ifdef CGAL_CARTESIAN_H
	p2 = Point((*pc_i).x(), (*pc_i).y());
#else
	p2 = Point((*pc_i).hx(), (*pc_i).hy(), (*pc_i).hw());
#endif
	
	if ((*pc_i).type()==Point_for_pm_bops<Point>::NORMAL_POINT){
	  rf_color = rf;
	  lf_color = lf;
	  e_color  = c;
	}
	if ((*pc_i).type()==Point_for_pm_bops<Point>::END_COMMON){
	  rf_color = (Color)((int)rf_color+(int)(*pc_i).color());
	}
	if (p1 != p2) 
        {
          all_segs.insert(Segment_for_pm_bops<Segment>(Segment(p1, p2),
						       e_color, 
						       lf_color, 
						       rf_color));
	}
	if ((*pc_i).type()==Point_for_pm_bops<Point>::BEGIN_COMMON){
	  lf_color = (Color)((int)(*pc_i).color()+(int)lf_color);
	  e_color  = (Color)3;
	}
	p1 = p2;
	
      }

      p2 = B;
      
      rf_color = rf;
      lf_color = lf;
      e_color  = c;
      if (p1 != p2) 
      {
        all_segs.insert(Segment_for_pm_bops<Segment>(Segment(p1, p2),   
						    c, lf, rf));
        p1 = p2;
      }
      return;
    }  
}; 
#endif

